@@ -81,8 +81,6 @@ class History:
 
     History objects should be created via a subclass of History.
 
-    Indexing
-    --------
     History acts like a sequence that can be indexed to return
     ``HistoryEntry`` objects.
 
@@ -125,6 +123,9 @@ class History:
         self.outs = None
         self.last_cmd_rtn = None
         self.last_cmd_out = None
+        self.hist_size = None
+        self.hist_units = None
+        self.remember_history = True
 
     def __len__(self):
         """Return the number of items in current session."""
@@ -208,6 +209,12 @@ class History:
         """
         pass
 
+    def clear(self):
+        """Clears the history of the current session from both the disk and
+        memory.
+        """
+        pass
+
 #
 # dummy
 #
@@ -250,53 +257,72 @@ cabc = _LazyModule.load('collections', 'collections.abc', 'cabc')
 xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
 xlj = _LazyModule.load('xonsh', 'xonsh.lazyjson', 'xlj')
 uptime = _LazyModule.load('xonsh', 'xonsh.xoreutils.uptime', 'uptime')
-def _xhj_gc_commands_to_rmfiles(hsize, files):
-    """Return the history files to remove to get under the command limit."""
-    rmfiles = []
+def _xhj_gc_commands_to_rmfiles(hsize, files) -> ([], int):
+    """Return number of units and list of history files to remove to get under the limit,
+
+    Parameters:
+    -----------
+    hsize (int):  units of history, # of commands in this case.
+    files ((mod_ts, num_commands, path)[], fsize): history files, sorted oldest first.
+
+    Returns:
+    --------
+    hsize_removed (int):  units of history to be removed
+    rm_files ((mod_ts, num_commands, path, fsize)[]): list of files to remove.
+    """
     n = 0
     ncmds = 0
-    for ts, fcmds, f in files[::-1]:
-        if fcmds == 0:
-            # we need to make sure that 'empty' history files don't hang around
-            rmfiles.append((ts, fcmds, f))
+    for _, fcmds, f, _ in reversed(files):
+        # `files` comes in with empty files included (now), don't need special handling to gc them here.
+
         if ncmds + fcmds > hsize:
             break
         ncmds += fcmds
         n += 1
-    rmfiles += files[:-n]
-    return rmfiles
+
+    cmds_removed = 0
+    files_removed = files[:-n]
+    for _, fcmds, f, _ in files_removed:
+        cmds_removed += fcmds
+
+    return cmds_removed, files_removed
 
 
 def _xhj_gc_files_to_rmfiles(hsize, files):
-    """Return the history files to remove to get under the file limit."""
+    """Return the number and list of history files to remove to get under the file limit."""
     rmfiles = files[:-hsize] if len(files) > hsize else []
-    return rmfiles
+    return len(rmfiles), rmfiles
 
 
 def _xhj_gc_seconds_to_rmfiles(hsize, files):
-    """Return the history files to remove to get under the age limit."""
-    rmfiles = []
+    """Return age removed (the age of the *oldest* file) and list of history files to remove to get under the age limit."""
     now = time.time()
-    for ts, _, f in files:
+    oldest_ts = files[0][0]
+    n = 0
+
+    for ts, _, f, _ in files:
         if (now - ts) < hsize:
             break
-        rmfiles.append((None, None, f))
-    return rmfiles
+        n += 1
+
+    return (now - oldest_ts) if n > 0 else 0, files[:n]
 
 
 def _xhj_gc_bytes_to_rmfiles(hsize, files):
     """Return the history files to remove to get under the byte limit."""
-    rmfiles = []
     n = 0
     nbytes = 0
-    for _, _, f in files[::-1]:
-        fsize = os.stat(f).st_size
+    for _, _, f, fsize in reversed(files):
         if nbytes + fsize > hsize:
             break
         nbytes += fsize
         n += 1
-    rmfiles = files[:-n]
-    return rmfiles
+    bytes_removed = 0
+    files_removed = files[:-n]
+    for _, _, f, fsize in files_removed:
+        bytes_removed += fsize
+
+    return bytes_removed, files_removed
 
 
 def _xhj_get_history_files(sort=True, newest_first=False):
@@ -323,7 +349,7 @@ def _xhj_get_history_files(sort=True, newest_first=False):
 class JsonHistoryGC(threading.Thread):
     """Shell history garbage collection."""
 
-    def __init__(self, wait_for_shell=True, size=None, *args, **kwargs):
+    def __init__(self, wait_for_shell=True, size=None, force=False, *args, **kwargs):
         """Thread responsible for garbage collecting old history.
 
         May wait for shell (and for xonshrc to have been loaded) to start work.
@@ -332,6 +358,7 @@ class JsonHistoryGC(threading.Thread):
         self.daemon = True
         self.size = size
         self.wait_for_shell = wait_for_shell
+        self.force_gc = force
         self.gc_units_to_rmfiles = {
             "commands": _xhj_gc_commands_to_rmfiles,
             "files": _xhj_gc_files_to_rmfiles,
@@ -344,6 +371,7 @@ class JsonHistoryGC(threading.Thread):
         while self.wait_for_shell:
             time.sleep(0.01)
         env = builtins.__xonsh__.env  # pylint: disable=no-member
+        xonsh_debug = env.get("XONSH_DEBUG", 0)
         if self.size is None:
             hsize, units = env.get("XONSH_HISTORY_SIZE")
         else:
@@ -353,34 +381,56 @@ class JsonHistoryGC(threading.Thread):
         if rmfiles_fn is None:
             raise ValueError("Units type {0!r} not understood".format(units))
 
-        for _, _, f in rmfiles_fn(hsize, files):
-            try:
-                os.remove(f)
-            except OSError:
-                pass
+        size_over, rm_files = rmfiles_fn(hsize, files)
+        hist = getattr(builtins.__xonsh__, "history", None)
+        if hist is not None:  # remember last gc pass history size
+            hist.hist_size = size_over + hsize
+            hist.hist_units = units
+
+        if self.force_gc or size_over < hsize:
+            i = 0
+            for _, _, f, _ in rm_files:
+                try:
+                    os.remove(f)
+                    if xonsh_debug:
+                        print(
+                            f"... Deleted {i:7d} of {len(rm_files):7d} history files.\r",
+                            end="",
+                        )
+                    pass
+                except OSError:
+                    pass
+                i += 1
+        else:
+            print(
+                f"Warning: History garbage collection would discard more history ({size_over} {units}) than it would keep ({hsize}).\n"
+                "Not removing any history for now. Either increase your limit ($XONSH_HISTORY_SIZE), or run `history gc --force`."
+            )
 
     def files(self, only_unlocked=False):
         """Find and return the history files. Optionally locked files may be
         excluded.
 
         This is sorted by the last closed time. Returns a list of
-        (timestamp, number of cmds, file name) tuples.
+        (file_size, timestamp, number of cmds, file name) tuples.
         """
-        # pylint: disable=no-member
-        env = getattr(builtins, "__xonsh__.env", None)
+        env = getattr(getattr(builtins, "__xonsh__", None), "env", None)
         if env is None:
             return []
+
+        xonsh_debug = env.get("XONSH_DEBUG", 0)
         boot = uptime.boottime()
         fs = _xhj_get_history_files(sort=False)
         files = []
         for f in fs:
             try:
-                if os.path.getsize(f) == 0:
+                cur_file_size = os.path.getsize(f)
+                if cur_file_size == 0:
                     # collect empty files (for gc)
-                    files.append((time.time(), 0, f))
+                    files.append((os.path.getmtime(f), 0, f, cur_file_size))
                     continue
                 lj = xlj.LazyJSON(f, reopen=False)
-                if lj["locked"] and lj["ts"][0] < boot:
+                if lj.get("locked", False) and lj["ts"][0] < boot:
                     # computer was rebooted between when this history was created
                     # and now and so this history should be unlocked.
                     hist = lj.load()
@@ -389,21 +439,30 @@ class JsonHistoryGC(threading.Thread):
                     with open(f, "w", newline="\n") as fp:
                         xlj.ljdump(hist, fp, sort_keys=True)
                     lj = xlj.LazyJSON(f, reopen=False)
-                if only_unlocked and lj["locked"]:
+                if only_unlocked and lj.get("locked", False):
                     continue
-                # info: closing timestamp, number of commands, filename
-                files.append((lj["ts"][1] or lj["ts"][0], len(lj.sizes["cmds"]) - 1, f))
+                # info: file size, closing timestamp, number of commands, filename
+                ts = lj.get("ts", (0.0, None))
+                files.append(
+                    (ts[1] or ts[0], len(lj.sizes["cmds"]) - 1, f, cur_file_size)
+                )
                 lj.close()
+                if xonsh_debug:
+                    print(f"... Enumerated {len(files):7d} history files.\r", end="")
             except (IOError, OSError, ValueError):
                 continue
-        files.sort()
+        files.sort()  # this sorts by elements of the tuple,
+        # the first of which just happens to be file mod time.
+        # so sort by oldest first.
         return files
 
 
 class JsonHistoryFlusher(threading.Thread):
     """Flush shell history to disk periodically."""
 
-    def __init__(self, filename, buffer, queue, cond, at_exit=False, *args, **kwargs):
+    def __init__(
+        self, filename, buffer, queue, cond, at_exit=False, skip=None, *args, **kwargs
+    ):
         """Thread for flushing history."""
         super(JsonHistoryFlusher, self).__init__(*args, **kwargs)
         self.filename = filename
@@ -412,6 +471,7 @@ class JsonHistoryFlusher(threading.Thread):
         queue.append(self)
         self.cond = cond
         self.at_exit = at_exit
+        self.skip = skip
         if at_exit:
             self.dump()
             queue.popleft()
@@ -436,9 +496,13 @@ class JsonHistoryFlusher(threading.Thread):
         for cmd in self.buffer:
             if "ignoredups" in opts and cmd["inp"] == last_inp:
                 # Skipping dup cmd
+                if self.skip is not None:
+                    self.skip(1)
                 continue
             if "ignoreerr" in opts and cmd["rtn"] != 0:
                 # Skipping failed cmd
+                if self.skip is not None:
+                    self.skip(1)
                 continue
             cmds.append(cmd)
             last_inp = cmd["inp"]
@@ -481,6 +545,9 @@ class JsonCommandField(cabc.Sequence):
         return len(self.hist)
 
     def __getitem__(self, key):
+        if not self.hist.remember_history:
+            return ""
+
         size = len(self)
         if isinstance(key, slice):
             return [self[i] for i in range(*key.indices(size))]
@@ -514,7 +581,7 @@ class JsonCommandField(cabc.Sequence):
 class JsonHistory(History):
     """Xonsh history backend implemented with JSON files.
 
-    JsonHistory implements two extra actions: ``diff``, and ``replay``.
+    JsonHistory implements an extra action: ``diff``
     """
 
     def __init__(self, filename=None, sessionid=None, buffersize=100, gc=True, **meta):
@@ -552,6 +619,7 @@ class JsonHistory(History):
         self._queue = collections.deque()
         self._cond = threading.Condition()
         self._len = 0
+        self._skipped = 0
         self.last_cmd_out = None
         self.last_cmd_rtn = None
         meta["cmds"] = []
@@ -566,7 +634,7 @@ class JsonHistory(History):
         self.rtns = JsonCommandField("rtn", self)
 
     def __len__(self):
-        return self._len
+        return self._len - self._skipped
 
     def append(self, cmd):
         """Appends command to history. Will periodically flush the history to file.
@@ -584,6 +652,8 @@ class JsonHistory(History):
         hf : JsonHistoryFlusher or None
             The thread that was spawned to flush history
         """
+        if not self.remember_history:
+            return
         self.buffer.append(cmd)
         self._len += 1  # must come before flushing
         if len(self.buffer) >= self.buffersize:
@@ -606,12 +676,22 @@ class JsonHistory(History):
         hf : JsonHistoryFlusher or None
             The thread that was spawned to flush history
         """
+        # Implicitly covers case of self.remember_history being False.
         if len(self.buffer) == 0:
             return
+
+        def skip(num):
+            self._skipped += num
+
         hf = JsonHistoryFlusher(
-            self.filename, tuple(self.buffer), self._queue, self._cond, at_exit=at_exit
+            self.filename,
+            tuple(self.buffer),
+            self._queue,
+            self._cond,
+            at_exit=at_exit,
+            skip=skip,
         )
-        self.buffer.clear()
+        self.buffer = []
         return hf
 
     def items(self, newest_first=False):
@@ -662,13 +742,29 @@ class JsonHistory(History):
         data["bufferlength"] = len(self.buffer)
         envs = builtins.__xonsh__.env
         data["gc options"] = envs.get("XONSH_HISTORY_SIZE")
+        data["gc_last_size"] = f"{(self.hist_size, self.hist_units)}"
         return data
 
-    def run_gc(self, size=None, blocking=True):
-        self.gc = JsonHistoryGC(wait_for_shell=False, size=size)
+    def run_gc(self, size=None, blocking=True, force=False):
+        self.gc = JsonHistoryGC(wait_for_shell=False, size=size, force=force)
         if blocking:
-            while self.gc.is_alive():
-                continue
+            while self.gc.is_alive():  # while waiting for gc.
+                time.sleep(0.1)  # don't monopolize the thread (or Python GIL?)
+
+    def clear(self):
+        """Clears the current session's history from both memory and disk."""
+
+        # Wipe history from memory. Keep sessionid and other metadata.
+        self.buffer = []
+        self.tss = JsonCommandField("ts", self)
+        self.inps = JsonCommandField("inp", self)
+        self.outs = JsonCommandField("out", self)
+        self.rtns = JsonCommandField("rtn", self)
+        self._len = 0
+        self._skipped = 0
+
+        # Flush empty history object to disk, overwriting previous data.
+        self.flush()
 
 #
 # sqlite
@@ -913,6 +1009,8 @@ class SqliteHistory(History):
         setattr(XH_SQLITE_CACHE, XH_SQLITE_CREATED_SQL_TBL, False)
 
     def append(self, cmd):
+        if not self.remember_history:
+            return
         envs = builtins.__xonsh__.env
         inp = cmd["inp"].rstrip()
         self.inps.append(inp)
@@ -970,6 +1068,21 @@ class SqliteHistory(History):
             while self.gc.is_alive():
                 continue
 
+    def clear(self):
+        """Clears the current session's history from both memory and disk."""
+        # Wipe memory
+        self.inps = []
+        self.rtns = []
+        self.outs = []
+        self.tss = []
+
+        # Wipe the current session's entries from the database.
+        sql = "DELETE FROM xonsh_history WHERE sessionid = ?"
+        with _xh_sqlite_get_conn(filename=self.filename) as conn:
+            c = conn.cursor()
+            _xh_sqlite_create_history_table(c)
+            c.execute(sql, (str(self.sessionid),))
+
 #
 # main
 #
@@ -982,6 +1095,7 @@ functools = _LazyModule.load('functools', 'functools')
 # amalgamated json
 # amalgamated os
 # amalgamated sys
+# amalgamated threading
 # amalgamated xonsh.history.base
 # amalgamated xonsh.history.dummy
 # amalgamated xonsh.history.json
@@ -1196,7 +1310,18 @@ def _XH_HISTORY_SESSIONS():
     }
 
 
-_XH_MAIN_ACTIONS = {"show", "id", "file", "info", "diff", "gc"}
+_XH_MAIN_ACTIONS = {
+    "show",
+    "id",
+    "file",
+    "info",
+    "diff",
+    "gc",
+    "flush",
+    "off",
+    "on",
+    "clear",
+}
 
 
 @functools.lru_cache()
@@ -1295,6 +1420,13 @@ def _xh_create_parser():
             'units; e.g. "--size 8128 commands"'
         ),
     )
+    gcp.add_argument(
+        "--force",
+        dest="force_gc",
+        default=False,
+        action="store_true",
+        help="perform garbage collection even if history much bigger than configured limit",
+    )
     bgcp = gcp.add_mutually_exclusive_group()
     bgcp.add_argument(
         "--blocking",
@@ -1316,11 +1448,19 @@ def _xh_create_parser():
         diff = subp.add_parser("diff", help="diff two xonsh history files")
         xdh.dh_create_parser(p=diff)
 
-        import xonsh.replay as xrp
+    # 'flush' subcommand
+    subp.add_parser("flush", help="flush the current history to disk")
 
-        replay = subp.add_parser("replay", help="replay a xonsh history file")
-        xrp.replay_create_parser(p=replay)
-        _XH_MAIN_ACTIONS.add("replay")
+    # 'off' subcommand
+    subp.add_parser("off", help="history will not be saved for this session")
+
+    # 'on' subcommand
+    subp.add_parser(
+        "on", help="history will be saved for the rest of the session (default)"
+    )
+
+    # 'clear' subcommand
+    subp.add_parser("clear", help="one-time wipe of session history")
 
     return p
 
@@ -1377,15 +1517,26 @@ def history_main(
             return
         print(str(hist.filename), file=stdout)
     elif ns.action == "gc":
-        hist.run_gc(size=ns.size, blocking=ns.blocking)
+        hist.run_gc(size=ns.size, blocking=ns.blocking, force=ns.force_gc)
     elif ns.action == "diff":
         if isinstance(hist, JsonHistory):
             xdh.dh_main_action(ns)
-    elif ns.action == "replay":
-        if isinstance(hist, JsonHistory):
-            import xonsh.replay as xrp
-
-            xrp.replay_main_action(hist, ns, stdout=stdout, stderr=stderr)
+    elif ns.action == "flush":
+        hf = hist.flush()
+        if isinstance(hf, threading.Thread):
+            hf.join()
+    elif ns.action == "off":
+        if hist.remember_history:
+            hist.clear()
+            hist.remember_history = False
+            print("History off", file=sys.stderr)
+    elif ns.action == "on":
+        if not hist.remember_history:
+            hist.remember_history = True
+            print("History on", file=sys.stderr)
+    elif ns.action == "clear":
+        hist.clear()
+        print("History cleared", file=sys.stderr)
     else:
         print("Unknown history action {}".format(ns.action), file=sys.stderr)
 
