@@ -396,13 +396,13 @@ def gitstatus_prompt():
         symbol = _get_def(category.upper())
         value = getattr(s, category)
         if symbol and value > 0:
-            ret += symbol + str(value) + "{NO_COLOR}"
+            ret += symbol + str(value) + "{RESET}"
     if s.staged + s.conflicts + s.changed + s.deleted + s.untracked + s.stashed == 0:
         symbol = _get_def("CLEAN")
         if symbol:
-            ret += symbol + "{NO_COLOR}"
+            ret += symbol + "{RESET}"
     ret = ret.rstrip("|")
-    ret += "{NO_COLOR}"
+    ret += "{RESET}"
 
     return ret
 
@@ -461,28 +461,20 @@ RE_REMOVE_ANSI = LazyObject(
 
 
 def _get_git_branch(q):
-    denv = builtins.__xonsh__.env.detype()
+    # create a safge detyped env dictionary and update with the additional git environment variables
+    # when running git status commands we do not want to acquire locks running command like git status
+    denv = dict(builtins.__xonsh__.env.detype())
+    denv.update({"GIT_OPTIONAL_LOCKS": "0"})
     try:
-        branches = xt.decode_bytes(
-            subprocess.check_output(
-                ["git", "branch"], env=denv, stderr=subprocess.DEVNULL
-            )
-        ).splitlines()
+        cmd = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
+        branch = xt.decode_bytes(
+            subprocess.check_output(cmd, env=denv, stderr=subprocess.DEVNULL)
+        )
+        branch = branch.splitlines()[0] or None
     except (subprocess.CalledProcessError, OSError, FileNotFoundError):
         q.put(None)
     else:
-        for branch in branches:
-            if not branch.startswith("* "):
-                continue
-            elif branch.endswith(")"):
-                branch = branch.split()[-1][:-1]
-            else:
-                branch = branch.split()[-1]
-
-            q.put(branch)
-            break
-        else:
-            q.put(None)
+        q.put(branch)
 
 
 def get_git_branch():
@@ -498,7 +490,6 @@ def get_git_branch():
     t.join(timeout=timeout)
     try:
         branch = q.get_nowait()
-        # branch = RE_REMOVE_ANSI.sub("", branch or "")
         if branch:
             branch = RE_REMOVE_ANSI.sub("", branch)
     except queue.Empty:
@@ -588,6 +579,15 @@ def _first_branch_timeout_message():
     )
 
 
+def _vc_has(binary):
+    """ This allows us to locate binaries after git only if necessary """
+    cmds = builtins.__xonsh__.commands_cache
+    if cmds.is_empty():
+        return bool(cmds.locate_binary(binary, ignore_alias=True))
+    else:
+        return bool(cmds.lazy_locate_binary(binary, ignore_alias=True))
+
+
 def current_branch():
     """Gets the branch for a current working directory. Returns an empty string
     if the cwd is not a repository.  This currently only works for git and hg
@@ -595,17 +595,9 @@ def current_branch():
     '<branch-timeout>' is returned.
     """
     branch = None
-    cmds = builtins.__xonsh__.commands_cache
-    # check for binary only once
-    if cmds.is_empty():
-        has_git = bool(cmds.locate_binary("git", ignore_alias=True))
-        has_hg = bool(cmds.locate_binary("hg", ignore_alias=True))
-    else:
-        has_git = bool(cmds.lazy_locate_binary("git", ignore_alias=True))
-        has_hg = bool(cmds.lazy_locate_binary("hg", ignore_alias=True))
-    if has_git:
+    if _vc_has("git"):
         branch = get_git_branch()
-    if not branch and has_hg:
+    if not branch and _vc_has("hg"):
         branch = get_hg_branch()
     if isinstance(branch, subprocess.TimeoutExpired):
         branch = "<branch-timeout>"
@@ -613,20 +605,43 @@ def current_branch():
     return branch or None
 
 
+def _get_exit_code(cmd):
+    """ Run a command and return its exit code """
+    denv = dict(builtins.__xonsh__.env.detype())
+    denv.update({"GIT_OPTIONAL_LOCKS": "0"})
+    child = subprocess.run(cmd, stderr=subprocess.DEVNULL, env=denv)
+    return child.returncode
+
+
 def _git_dirty_working_directory(q, include_untracked):
-    status = None
-    denv = builtins.__xonsh__.env.detype()
     try:
-        cmd = ["git", "status", "--porcelain"]
+        # Borrowed from this conversation
+        # https://gist.github.com/sindresorhus/3898739
         if include_untracked:
-            cmd.append("--untracked-files=normal")
+            cmd = [
+                "git",
+                "status",
+                "--porcelain",
+                "--quiet",
+                "--untracked-files=normal",
+            ]
+            exitcode = _get_exit_code(cmd)
         else:
-            cmd.append("--untracked-files=no")
-        status = subprocess.check_output(cmd, stderr=subprocess.DEVNULL, env=denv)
+            # checking unindexed files is faster, so try that first
+            unindexed = ["git", "diff-files", "--quiet"]
+            exitcode = _get_exit_code(unindexed)
+            if exitcode == 0:
+                # then, check indexed files
+                indexed = ["git", "diff-index", "--quiet", "--cached", "HEAD"]
+                exitcode = _get_exit_code(indexed)
+        # "--quiet" git commands imply "--exit-code", which returns:
+        # 1 if there are differences
+        # 0 if there are no differences
+        dwd = bool(exitcode)
     except (subprocess.CalledProcessError, OSError, FileNotFoundError):
         q.put(None)
-    if status is not None:
-        return q.put(bool(status))
+    else:
+        q.put(dwd)
 
 
 def git_dirty_working_directory(include_untracked=False):
@@ -680,10 +695,9 @@ def dirty_working_directory():
     None. Currently supports git and hg.
     """
     dwd = None
-    cmds = builtins.__xonsh__.commands_cache
-    if cmds.lazy_locate_binary("git", ignore_alias=True):
+    if _vc_has("git"):
         dwd = git_dirty_working_directory()
-    if cmds.lazy_locate_binary("hg", ignore_alias=True) and dwd is None:
+    if dwd is None and _vc_has("hg"):
         dwd = hg_dirty_working_directory()
     return dwd
 
@@ -750,54 +764,60 @@ class PromptFormatter:
     def __init__(self):
         self.cache = {}
 
-    def __call__(self, template=DEFAULT_PROMPT, fields=None):
+    def __call__(self, template=DEFAULT_PROMPT, fields=None, **kwargs):
         """Formats a xonsh prompt template string."""
         if fields is None:
             self.fields = builtins.__xonsh__.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
         else:
             self.fields = fields
         try:
-            prompt = self._format_prompt(template=template)
+            prompt = self._format_prompt(template=template, **kwargs)
         except Exception:
             return _failover_template_format(template)
         # keep cache only during building prompt
         self.cache.clear()
         return prompt
 
-    def _format_prompt(self, template=DEFAULT_PROMPT):
+    def _format_prompt(self, template=DEFAULT_PROMPT, **kwargs):
+        return "".join(self._get_tokens(template, **kwargs))
+
+    def _get_tokens(self, template, **kwargs):
         template = template() if callable(template) else template
         toks = []
         for literal, field, spec, conv in xt.FORMATTER.parse(template):
             toks.append(literal)
-            entry = self._format_field(field, spec, conv)
+            entry = self._format_field(field, spec, conv, idx=len(toks), **kwargs)
             if entry is not None:
                 toks.append(entry)
-        return "".join(toks)
+        return toks
 
-    def _format_field(self, field, spec, conv):
+    def _format_field(self, field, spec, conv, **kwargs):
         if field is None:
             return
         elif field.startswith("$"):
             val = builtins.__xonsh__.env[field[1:]]
             return _format_value(val, spec, conv)
         elif field in self.fields:
-            val = self._get_field_value(field)
+            val = self._get_field_value(field, **kwargs)
             return _format_value(val, spec, conv)
         else:
             # color or unknown field, return as is
             return "{" + field + "}"
 
-    def _get_field_value(self, field):
+    def _get_field_value(self, field, **kwargs):
         field_value = self.fields[field]
         if field_value in self.cache:
             return self.cache[field_value]
+        return self._no_cache_field_value(field, field_value, **kwargs)
+
+    def _no_cache_field_value(self, field, field_value, **_):
         try:
             value = field_value() if callable(field_value) else field_value
             self.cache[field_value] = value
         except Exception:
             print("prompt: error: on field {!r}" "".format(field), file=sys.stderr)
             xt.print_exception()
-            value = "{{BACKGROUND_RED}}{{ERROR:{}}}{{NO_COLOR}}".format(field)
+            value = "{{BACKGROUND_RED}}{{ERROR:{}}}{{RESET}}".format(field)
         return value
 
 
@@ -831,21 +851,21 @@ def default_prompt():
         dp = (
             "{env_name}"
             "{BOLD_GREEN}{user}@{hostname}"
-            "{BOLD_BLUE} {cwd} {prompt_end}{NO_COLOR} "
+            "{BOLD_BLUE} {cwd} {prompt_end}{RESET} "
         )
     elif xp.ON_WINDOWS and not xp.win_ansi_support():
         dp = (
             "{env_name}"
             "{BOLD_INTENSE_GREEN}{user}@{hostname}{BOLD_INTENSE_CYAN} "
-            "{cwd}{branch_color}{curr_branch: {}}{NO_COLOR} "
-            "{BOLD_INTENSE_CYAN}{prompt_end}{NO_COLOR} "
+            "{cwd}{branch_color}{curr_branch: {}}{RESET} "
+            "{BOLD_INTENSE_CYAN}{prompt_end}{RESET} "
         )
     else:
         dp = (
             "{env_name}"
             "{BOLD_GREEN}{user}@{hostname}{BOLD_BLUE} "
-            "{cwd}{branch_color}{curr_branch: {}}{NO_COLOR} "
-            "{BOLD_BLUE}{prompt_end}{NO_COLOR} "
+            "{cwd}{branch_color}{curr_branch: {}}{RESET} "
+            "{BOLD_BLUE}{prompt_end}{RESET} "
         )
     return dp
 
@@ -893,7 +913,7 @@ def multiline_prompt(curr=""):
             basetoks.append(("\001" + pre + "\002", post))
             baselen += len(post)
     if baselen == 0:
-        return xt.format_color("{NO_COLOR}" + tail, hide=True)
+        return xt.format_color("{RESET}" + tail, hide=True)
     toks = basetoks * (headlen // baselen)
     n = headlen % baselen
     count = 0
@@ -909,7 +929,7 @@ def multiline_prompt(curr=""):
         count = newcount
         if n <= count:
             break
-    toks.append((xt.format_color("{NO_COLOR}", hide=True), tail))
+    toks.append((xt.format_color("{RESET}", hide=True), tail))
     rtn = "".join(itertools.chain.from_iterable(toks))
     return rtn
 
